#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the resource
#   module builder playbook.
#
# Do not edit this file manually.
#
# Changes to this file will be over written
#   by the resource module builder.
#
# Changes should be made in the model used to
#   generate this file or in the resource module
#   builder template.
#
#############################################
"""
The arg spec for the vyos_lldp_interfaces module
"""

from __future__ import annotations


class Lldp_interfacesArgs(object):  # pylint: disable=R0903
    """The arg spec for the vyos_lldp_interfaces module
    """

    def __init__(self, **kwargs):
        pass

    argument_spec = {
        "config": {
            "elements": "dict",
            "options": {
                "enable": {"default": True, "type": "bool"},
                "location": {
                    "options": {
                        "civic_based": {
                            "options": {
                                "ca_info": {
                                    "elements": "dict",
                                    "options": {
                                        "ca_type": {"type": "int"},
                                        "ca_value": {"type": "str"},
                                    },
                                    "type": "list",
                                },
                                "country_code": {
                                    "required": True,
                                    "type": "str",
                                },
                            },
                            "type": "dict",
                        },
                        "coordinate_based": {
                            "options": {
                                "altitude": {"type": "int"},
                                "datum": {
                                    "choices": ["WGS84", "NAD83", "MLLW"],
                                    "type": "str",
                                },
                                "latitude": {"required": True, "type": "str"},
                                "longitude": {"required": True, "type": "str"},
                            },
                            "type": "dict",
                        },
                        "elin": {"type": "str"},
                    },
                    "type": "dict",
                },
                "name": {"required": True, "type": "str"},
            },
            "type": "list",
        },
        "state": {
            "choices": ["merged", "replaced", "overridden", "deleted"],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301
