# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Ansible, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations


class ModuleDocFragment(object):

    # Standard documentation fragment
    DOCUMENTATION = r'''
options:
  backup:
    description:
    - Create a backup file including the timestamp information so you can get
      the original file back if you somehow clobbered it incorrectly.
    type: bool
    default: no
'''
